/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* This software is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
\*************************************************************************/

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include <cadef.h>
#include <alarm.h>
#include <sys/stat.h>

#include "flashBurner.h"

struct fileInfo {
    const char   *name;
    FILE         *fp;
    unsigned int  offset;
    off_t         bytesRead;
    int           done;
    int           forceBinary;
    unsigned long ihex_ulba;    /* Intel hex format upper linear base address */
    enum          { format_binary, format_srec, format_ihex } format;
};

/*
 * S-record support routines
 */
static int
mustGetc(struct fileInfo *fp)
{
    int c = getc(fp->fp);

    if (c == EOF) {
        fprintf(stderr, "Unexpected EOF while reading %s\n", fp->name);
        exit(1);
    }
    fp->bytesRead++;
    return c;
}

static int
getHexChar(struct fileInfo *fp)
{
    switch (mustGetc(fp)) {
    case '0':           return 0x0;
    case '1':           return 0x1;
    case '2':           return 0x2;
    case '3':           return 0x3;
    case '4':           return 0x4;
    case '5':           return 0x5;
    case '6':           return 0x6;
    case '7':           return 0x7;
    case '8':           return 0x8;
    case '9':           return 0x9;
    case 'A': case 'a': return 0xA;
    case 'B': case 'b': return 0xB;
    case 'C': case 'c': return 0xC;
    case 'D': case 'd': return 0xD;
    case 'E': case 'e': return 0xE;
    case 'F': case 'f': return 0xF;
    default:
        fprintf(stderr, "Unexpected EOF while reading %s\n", fp->name);
        exit(1);
    }
}

static int
getHexPair(struct fileInfo *fp, int *checksum)
{
    int h, l, p;

    h = getHexChar(fp);
    l = getHexChar(fp);
    p = (h << 4) + l;
    *checksum = (*checksum + p) & 0xFF;
    return p;
}

/*
 * Read input file
 */
static int
fillBuffer(struct flashBurnerWaveformData *wdp, struct fileInfo *fp)
{
    if (fp->format == format_srec) {
        long backup;
        off_t saveBytesRead;
        int byteCount = 0;
        int sByteCount, nAddr;
        unsigned long addr;
        int recType, checksum;
        int lastAddr = 0;
        int firstTime = 1;
        int c;

        if (fp->done)
            return 0;
        for (;;) {
            backup = ftell(fp->fp);
            saveBytesRead = fp->bytesRead;
            do {
                c = getc(fp->fp);
                if (c == EOF) {
                    fp->done = 1;
                    return byteCount;
                }
                fp->bytesRead++;
            } while (c == '\r');
            if (c != 'S') {
                fprintf(stderr, "\"%s\" line does not begin with 'S'.\n", fp->name);
                exit(1);
            }
            switch((recType = mustGetc(fp))) {
            case '0':   nAddr = 2; break;
            case '1':   nAddr = 2; break;
            case '2':   nAddr = 3; break;
            case '3':   nAddr = 4; break;
            case '7':   nAddr = 2; fp->done = 1; break;
            case '8':   nAddr = 3; fp->done = 1; break;
            case '9':   nAddr = 4; fp->done = 1; break;
            default:
                fprintf(stderr, "\"%s\" has bad S-record type.\n", fp->name);
                exit(1);
            }
            checksum = 1;
            sByteCount = getHexPair(fp, &checksum);
            if (sByteCount < (nAddr + 1)) {
                fprintf(stderr, "\"%s\" S-record byte count too small.\n", fp->name);
                exit(1);
            }
            sByteCount -= nAddr + 1;
            addr = 0;
            while (nAddr--)
                addr = (addr << 8) | getHexPair(fp, &checksum);
            if ((recType < '1') || (recType > '3')) {
                while (sByteCount--)
                    getHexPair(fp, &checksum);
            }
            else {
                if (firstTime) {
                    firstTime = 0;
                    lastAddr = addr;
                    fp->offset = addr;
                }
                if ((addr != lastAddr)
                 || ((byteCount + sByteCount) > sizeof wdp->flashData)) {
                    fseek(fp->fp, backup, SEEK_SET);
                    fp->bytesRead = saveBytesRead;
                    return byteCount;
                }
                lastAddr = addr + sByteCount;
                while (sByteCount--)
                    wdp->flashData[byteCount++] = getHexPair(fp, &checksum);
            }
            getHexPair(fp, &checksum);
            if (checksum != 0) {
                fprintf(stderr, "\"%s\" S-record bad checksum.\n", fp->name);
                exit(1);
            }
            while (mustGetc(fp) != '\n')
                continue;
        }
    }
    else if (fp->format == format_ihex) {
        long backup;
        off_t saveBytesRead;
        int byteCount = 0;
        int recLen, loadOffset;
        unsigned long addr;
        int recType, checksum;
        int lastAddr = 0;
        int firstTime = 1;
        int c;

        if (fp->done)
            return 0;
        for (;;) {
            backup = ftell(fp->fp);
            saveBytesRead = fp->bytesRead;
            do {
                c = getc(fp->fp);
                if (c == EOF) {
                    fp->done = 1;
                    return byteCount;
                }
                fp->bytesRead++;
            } while (c == '\r');
            if (c != ':') {
                fprintf(stderr, "\"%s\" line does not begin with ':'.\n", fp->name);
                exit(1);
            }
            checksum = 0;
            recLen = getHexPair(fp, &checksum);
            loadOffset = getHexPair(fp, &checksum) << 8;
            loadOffset |= getHexPair(fp, &checksum);
            recType = getHexPair(fp, &checksum);
            if (recType == 0x04) {
                if ((recLen != 2)
                 || (loadOffset != 0)) {
                    fprintf(stderr, "\"%s\" garbled Extended Linear Address Record.\n", fp->name);
                    exit(1);

                 }
                fp->ihex_ulba = getHexPair(fp, &checksum) << 8;
                fp->ihex_ulba |= getHexPair(fp, &checksum);
            }
            else if (recType == 0x01) {
                if ((recLen != 0)
                 || (loadOffset != 0)) {
                    fprintf(stderr, "\"%s\" garbled End of File Record.\n", fp->name);
                    exit(1);
                }
                fp->done = 1;
            }
            else if (recType == 0x00) {
                addr = fp->ihex_ulba << 16 | loadOffset;
                if (firstTime) {
                    firstTime = 0;
                    lastAddr = addr;
                    fp->offset = addr;
                }
                if ((addr != lastAddr)
                 || ((byteCount + recLen) > sizeof wdp->flashData)) {
                    fseek(fp->fp, backup, SEEK_SET);
                    fp->bytesRead = saveBytesRead;
                    return byteCount;
                }
                lastAddr = addr + recLen;
                while (recLen--)
                    wdp->flashData[byteCount++] = getHexPair(fp, &checksum);
            }
            else {
                fprintf(stderr, "\"%s\" unexpected record type %#x.\n", fp->name, recType);
                exit(1);
            }
            getHexPair(fp, &checksum);
            if (checksum != 0) {
                fprintf(stderr, "\"%s\" Bad checksum.\n", fp->name);
                exit(1);
            }
            while (mustGetc(fp) != '\n')
                continue;
        }
    }
    else {
        int byteCount = fread(wdp->flashData, 1, sizeof wdp->flashData, fp->fp);
        if (ferror(fp->fp)) {
            fprintf(stderr, "Error reading \"%s\": %s\n", fp->name,
                                                              strerror(errno));
            exit(1);
        }
        if ((fp->forceBinary == 0)
         && (byteCount > 6)
         && (fp->bytesRead == 0)) {
            if ((wdp->flashData[0] == 'S')
             && (wdp->flashData[1] >= '0')
             && (wdp->flashData[1] <= '3')
             && isxdigit(wdp->flashData[2])
             && isxdigit(wdp->flashData[3])) {
                rewind(fp->fp);
                fp->format = format_srec;
                return fillBuffer(wdp, fp);
            }
            if ((wdp->flashData[0] == ':')
             && isxdigit(wdp->flashData[1])
             && isxdigit(wdp->flashData[2])
             && isxdigit(wdp->flashData[3])
             && isxdigit(wdp->flashData[4])
             && isxdigit(wdp->flashData[5])) {
                rewind(fp->fp);
                fp->format = format_ihex;
                return fillBuffer(wdp, fp);
            }
        }
        fp->forceBinary = 1;
        fp->bytesRead += byteCount;
        return byteCount;
    }
}

/*
 * Chanel access I/O-completion callback
 */
static void
callback(struct event_handler_args args)
{
    int ret = 0;

    if (args.status != ECA_NORMAL) {
        fprintf(stderr, "I/O operation failed: %s\n", ca_message(args.status));
        ret = -1;
    }
    *(int *)args.usr = ret;
}

void
printProgressBar(char *msg, int pct)
{
    int nHashes = (50*pct/100);
    int i;

    printf("\r%s [", msg);
    for (i=0; i<nHashes; i++)
        putchar('#');
    while (i++ < 50)
        putchar('.');
    printf ("] %3d%%", pct);
    fflush(stdout);
}

int main(int argc,char **argv)
{
    struct fileInfo fileInfo, *fp = &fileInfo;
    int busy;
    struct flashBurnerWaveformData wd;
    int byteCount;
    chid mychid;
    int eca;
    struct dbr_sts_int status;
    const char *progName = "flashBurner";
    const char *pvName;
    int hFlag = 0, qFlag = 0, rFlag = 0;
    int c;
    struct stat sb;
    int percentTransferred = 0;
    char statusMessage[80];
    extern int getopt(int argc, char * const argv[], const char *optstring);
    extern int optind;

    /*
     * Process arguments
     */
    fp->forceBinary = 0;
    fp->format = format_binary;
    if (argc) {
        if ((progName = strrchr(argv[0], '/')) == NULL)
            progName = argv[0];
        else
            progName++;
    }
    while ((c = getopt(argc, argv, "bhqr")) != EOF) {
        switch(c) {
        case 'b':   fp->forceBinary = 1;  break;
        case 'h':   hFlag = 1;  break;
        case 'q':   qFlag = 1;  break;
        case 'r':   rFlag = 1;  break;
        default:    hFlag = 1;  break;
        }
    }
    if ((argc - optind) != 2)
        hFlag = 1;
    if (hFlag) {
        fprintf(stderr, "Usage: %s [-b] [-h] [-q] [-r] pvName fIlename\n", progName);
        fprintf(stderr, "      -b  Force binary programming even if file appears to contain Intel Hex or S-records\n");
        fprintf(stderr, "      -h  Display this message\n");
        fprintf(stderr, "      -q  Quiet (no download status updates)\n");
        fprintf(stderr, "      -r  Bit reverse bytes before programming (.rpd files, EPCS devices)\n");
        return 1;
    }
    pvName = argv[optind];
    fp->name = argv[optind+1];

    /*
     * Open connections to file and PV
     */
    if ((fp->fp = fopen(fp->name, "r")) == NULL) {
        fprintf(stderr, "%s: Can't open \"%s\": %s\n", progName,
                                                       fp->name,
                                                       strerror(errno));
        return 1;
    }
    eca = ca_context_create(ca_disable_preemptive_callback);
    if (eca != ECA_NORMAL) {
        fprintf(stderr, "ca_context_create failed: %s.\n", ca_message(eca));
        return 2;
    }
    eca = ca_create_channel(pvName, NULL, NULL, 10, &mychid);
    if (eca != ECA_NORMAL) {
        fprintf(stderr, "ca_create_channel failed: %s.\n", ca_message(eca));
        return 2;
    }
    eca = ca_pend_io(8.0);
    if (eca != ECA_NORMAL) {
        if (eca == ECA_TIMEOUT)
            fprintf(stderr, "ca_pend_io timed out -- likely because "
                            "\"%s\" is an invalid PV name or its IOC "
                            "is unreachable.\n", pvName);
        else
            fprintf(stderr, "ca_pend_io failure: %s.\n", ca_message(eca));
        return 2;
    }

    /*
     * Get file size
     */
    if (fstat(fileno(fp->fp), &sb) < 0) {
        fprintf(stderr, "Can't determine size of \"%s\".\n", fp->name);
        return 1;
    }

    /*
     * Transfer the image
     */
    fp->bytesRead = 0;
    fp->done = 0;
    fp->offset = 0;
    byteCount = 0;
    for (;;) {
        int i;
        unsigned int checksum = 0;
        for (i = 0 ; i < FLASH_PAGE_SIZE ; i++)
            checksum += (unsigned char)wd.flashData[i];
        sprintf(wd.header, FLASH_HEADER_FORMAT, byteCount, fp->offset, checksum);
        busy = 1;
        eca = ca_array_put_callback(DBR_CHAR,
                                    sizeof wd,
                                    mychid,
                                    (void *)&wd,
                                    callback,
                                    (void *)&busy);
        if (eca != ECA_NORMAL) {
            fprintf(stderr, "ca_put_array_callback failed:%s.\n", ca_message(eca));
            return 2;
        }
        while (busy > 0) {
            eca = ca_pend_event(0.1);
            if ((eca != ECA_NORMAL) && (eca != ECA_TIMEOUT)) {
                if (!qFlag) printf("\n");
                fprintf(stderr, "ca_pend_event failed: %s.\n", ca_message(eca));
                return 2;
            }
            if (!qFlag) {
                percentTransferred = (fp->bytesRead*100.0)/sb.st_size;
                sprintf(statusMessage, "\r%srogramming:",
                                    fp->format == format_srec ? "S-record p" : 
                                    fp->format == format_ihex ? "IntelHex p" :
                                                                "P"); 
                printProgressBar(statusMessage, percentTransferred);
            }
        }
        if (busy < 0)
            return 2;
        eca = ca_get(DBR_STS_INT, mychid, &status);
        if (eca != ECA_NORMAL) {
            fprintf(stderr, "ca_get failed: %s.\n", ca_message(eca));
            return 2;
        }
        eca = ca_pend_io(5.0);
        if (eca != ECA_NORMAL) {
            fprintf(stderr, "ca_pend_io failure: %s.\n", ca_message(eca));
            return 2;
        }
        if (status.severity != NO_ALARM) {
            if (!qFlag) printf("\n");
            printf("Programming failed -- record is in alarm state.\n");
            return 2;
        }
        fp->offset += byteCount;
        byteCount = fillBuffer(&wd, fp);
        if (byteCount == 0)
            break;
        if (rFlag) {
            int i;
            for (i = 0 ; i < byteCount ; i++) {
                int b;
                char bn = wd.flashData[i], br = 0;
                for (b = 0 ; b < 8 ; b++) {
                    br <<= 1;
                    if (bn & 0x1)
                        br |= 1;
                    bn >>= 1;
                }
                wd.flashData[i] = br;
            }
        }
    }
    if (!qFlag)
        printf("\n");
    fclose(fp->fp);
    return(0);
}
